from h5py import File
import logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)
import pygid
from typing import Any
from dataclasses import dataclass
def save_fit(filename, entry, img_container_fit, frame_num):
    with File(filename, "r+") as f:
        group_name = f"/{entry}/data/analysis/frame{str(frame_num).zfill(5)}"
        pygid._save_img_container_fit(f, group_name, img_container_fit)
        logger.info(f"Saved fitted peaks to file: {filename}, entry: {entry}, frame: {frame_num}")
        return


@dataclass
class DetectedPeaks:
    radius: Any = None
    radius_width: Any = None
    angle: Any = None
    angle_width: Any = None

    def __post_init__(self):
        pass

def read_detected_peaks(nexus, entry, frame_num):
    group_name = f"/{entry}/data/analysis/frame{str(frame_num).zfill(5)}"
    entry_dict = nexus.entry_dict
    if not entry in entry_dict:
        raise KeyError(f"entry {entry} not in the file. The file structure: {entry_dict}")
    return nexus.get_dataset(f"{group_name}/detected_peaks")




import numpy as np
import cv2
from typing import Tuple, Any, Optional
import importlib.metadata
import logging
from datetime import datetime

logger = logging.getLogger()
logger.setLevel(logging.INFO)

for lib in ["matplotlib" , "numba", "h5py"]:
    logging.getLogger(lib).setLevel(logging.WARNING)
import time
from dataclasses import dataclass
import pygid


from pygidfit.fitting_models import (
    fit_clusters_multiprocessing,
    fit_ring_cluster,
    fit_peak_on_ring_cluster,
    fit_peak_cluster,
)

from pygidfit.box_utils import (
    boxes_preprocessing,
    make_box_attributes,
)

from pygidfit.io_utils import (
    save_fit,
    DetectedPeaks,
    read_detected_peaks,
)

from pygidfit.imgcontainer import ImageContainer
from pygidfit.clustering_and_errors import cluster_boxes_by_centers

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.patches import Rectangle

def calc_smpl_hor(ai, crit_angle, wavelength):
    """
    Calculate the critical scattering vector (q) for the sample horizon.

    Parameters
    ----------
    ai : float
        Incident angle of the beam (degrees).
    crit_angle : float
        Critical angle of the sample (degrees).
    wavelength : float
        Wavelength of the radiation (meters).

    Returns
    -------
    float
        Critical q value corresponding to the sample horizon (Å⁻¹).
    """
    return np.pi * 4 / wavelength * np.sin(np.deg2rad((ai + crit_angle)) / 2) / 10


def img_preprocessing(img,  ai, crit_angle, wavelength, q_z):
    """
    Preprocess a reciprocal-space image by masking below the critical q_z
    and non-positive pixels.

    Parameters
    ----------
    img : np.ndarray
        2D reciprocal-space image to preprocess.
    ai : float
        Incident angle of the beam (degrees).
    crit_angle : float
        Critical angle of the sample (degrees).
    wavelength : float
        Radiation wavelength (meters).
    q_z : np.ndarray
        Out-of-plane scattering vector array corresponding to the image.

    Returns
    -------
    np.ndarray
        Preprocessed image with masked pixels replaced by NaN.
    """
    q_z_critical = calc_smpl_hor(ai, crit_angle, wavelength)
    mask = q_z < q_z_critical
    img[mask, :] = np.nan
    img[img <= 0] = np.nan
    return img


def _get_polar_grid(
        img_shape: Tuple[int, int],
        polar_shape: Tuple[int, int],
        beam_center: Tuple[float, float],
) -> Tuple[np.ndarray, np.ndarray, float]:
    """
    Generate a polar coordinate grid for a 2D image.

    Parameters
    ----------
    img_shape : tuple of int
        Shape of the original 2D image (rows, columns).
    polar_shape : tuple of int
        Shape of the polar grid to generate (polar angles, radial bins).
    beam_center : tuple of float
        Coordinates of the beam center (y0, z0) in the image.

    Returns
    -------
    polar_yy : np.ndarray
        Y-coordinates of the polar grid in Cartesian space.
    polar_zz : np.ndarray
        Z-coordinates of the polar grid in Cartesian space.
    phi_max_deg : float
        Maximum polar angle in degrees.
    """
    y0, z0 = beam_center

    y = (np.arange(img_shape[1]) - y0)
    z = (np.arange(img_shape[0]) - z0)

    yy, zz = np.meshgrid(y, z)
    rr = np.sqrt(yy ** 2 + zz ** 2)
    phi = np.arctan2(zz, yy)
    r_range = (rr.min(), rr.max())
    phi_range = phi.min(), phi.max()

    phi = np.linspace(*phi_range, polar_shape[0])
    r = np.linspace(*r_range, polar_shape[1])

    r_matrix = r[np.newaxis, :].repeat(polar_shape[0], axis=0)
    p_matrix = phi[:, np.newaxis].repeat(polar_shape[1], axis=1)

    polar_yy = r_matrix * np.cos(p_matrix) + y0
    polar_zz = r_matrix * np.sin(p_matrix) + z0

    return polar_yy, polar_zz, np.rad2deg(phi.max())

def polar_conversion(img: np.ndarray, yy: np.ndarray, zz: np.ndarray, algorithm: int) -> np.ndarray or None:
    """
    Convert a 2D image to polar coordinates using a remapping grid.

    Parameters
    ----------
    img : np.ndarray
        Original 2D image to convert.
    yy : np.ndarray
        X-coordinate grid for remapping (from `_get_polar_grid`).
    zz : np.ndarray
        Y-coordinate grid for remapping (from `_get_polar_grid`).
    algorithm : int
        Interpolation method (e.g., cv2.INTER_LINEAR, cv2.INTER_NEAREST).

    Returns
    -------
    np.ndarray
        Polar-transformed 2D image.
    """

    return cv2.remap(img.astype(np.float32),
                        yy.astype(np.float32),
                        zz.astype(np.float32),
                        interpolation=algorithm,
                        borderValue=np.nan)



def show_masked_images_debug(img, masked_img, boxes, clusters, debug=True):
    """
    Display images with boxes and print example data if debug is True.

    Parameters
    ----------
    img : np.ndarray
        Original image.
    masked_img : np.ndarray
        Masked version of the image.
    boxes : list
        List of box objects with a 'limits' attribute (xmin, ymin, xmax, ymax).
    clusters : list
        List of cluster objects.
    debug : bool
        If True, display plots and print examples.
    """
    if not debug:
        return

    # Show original image with boxes
    fig, axes = plt.subplots(figsize=(6, 6))
    norm = LogNorm(vmin=np.nanmin(img[img > 0]), vmax=np.nanmax(img))
    axes.imshow(img, cmap='inferno', origin='lower', norm=norm)

    for box in boxes:
        xmin, ymin, xmax, ymax = map(int, box.limits)
        width = xmax - xmin
        height = ymax - ymin

        rect = Rectangle(
            (xmin, ymin), width, height,
            linewidth=1.5, edgecolor='red', facecolor='none'
        )
        axes.add_patch(rect)

    plt.show()

    # Show masked image
    fig, axes = plt.subplots(figsize=(6, 6))
    norm = LogNorm(vmin=np.nanmin(masked_img[masked_img > 0]), vmax=np.nanmax(masked_img))
    axes.imshow(masked_img, cmap='inferno', origin='lower', norm=norm)
    axes.set_title('masked image')
    plt.show()

    # Print example data
    print("box example: ", boxes[0], "cluster example: ", clusters[0])


def fit_single_image(polar_img, boxes, clusters, theta_fixed=True, peaks_pool=None, debug=False,
                     multiprocessing=False):
    """
    Fit clusters in a polar-transformed image with Gaussian models.

    This function applies Gaussian fitting to clusters of peaks, rings, or combined
    peak-on-ring clusters. It optionally masks irrelevant areas of the image.

    Parameters
    ----------
    polar_img : ndarray
        2D polar-transformed scattering image (axis 0: polar angle, axis 1: radial coordinate |q|).
    boxes : list
        List of `Boxes` objects containing detected peak information.
    clusters : list
        List of cluster objects, each representing peaks, rings, or both.
    theta_fixed : bool, optional
        Fix Gaussian tilt angle to 0° (azimuthal direction) during fitting. Default is True.
    peaks_pool : list or None, optional
        Pool of previously fitted peaks to reuse parameters for sequential frames. Default is None.
    debug : bool, optional
        Enable debug mode to visualize masked images and fitting results. Default is False.
    multiprocessing : bool, optional
        Enable multiprocessing for cluster fitting. Default is False.

    Notes
    -----
    - Clusters of type 'peaks' and 'both' are masked before fitting.
    - Depending on cluster type, the appropriate fitting function is called:
        - 'peaks' → `fit_peak_cluster`
        - 'rings' → `fit_ring_cluster`
        - 'both' → `fit_peak_on_ring_cluster`
    - Fitted parameters are stored back into the `boxes` objects.
    - If debug mode is enabled, execution time is printed.
    """
    mask = np.zeros_like(polar_img, dtype=bool)

    for cluster in clusters:
        if cluster.type == 'peaks' or cluster.type == 'both':
            xmin, ymin, xmax, ymax = map(int, cluster.bbox)
            mask[ymin:ymax, xmin:xmax] = True

    masked_img = np.where(~mask, polar_img, np.nan)

    show_masked_images_debug(polar_img, masked_img, boxes, clusters, debug=debug)

    time0 = time.time()

    if multiprocessing:
        fit_clusters_multiprocessing(clusters, boxes, polar_img, masked_img, theta_fixed, debug=debug)
    else:
        for cluster in clusters:
            if cluster.type == 'rings':
                fitting_result = fit_ring_cluster(cluster, boxes, masked_img, peaks_pool, debug)
                make_box_attributes(cluster.indices, boxes, fitting_result, cluster.type, debug)
            elif cluster.type == 'peaks':
                fitting_result = fit_peak_cluster(cluster, boxes, polar_img, peaks_pool, theta_fixed, debug)
                make_box_attributes(cluster.indices, boxes, fitting_result, cluster.type, debug)
        for cluster in clusters:
            if cluster.type == 'both':
                fitting_result = fit_peak_on_ring_cluster(cluster, boxes, polar_img, peaks_pool, theta_fixed, debug)
                make_box_attributes(cluster.indices, boxes, fitting_result, cluster.type, debug)

    time1 = time.time()
    if debug:
        print(f"image fitting took {(time1 - time0) * 1000} ms")




def compute_qzqxy_with_error(radius, sigma_radius, angle_deg, sigma_angle_deg, qz_max, qxy_max):
    """
    Computes qx and qy from radius and angle (in degrees) and propagates errors.

    Returns:
        q_array: np.array([qx, qy])
        sigma_array: np.array([sigma_qx, sigma_qy])
    """
    theta = np.deg2rad(angle_deg)
    sigma_theta = np.deg2rad(sigma_angle_deg)

    # coordinates
    qz = radius * np.sin(theta)
    qxy = radius * np.cos(theta)

    # error propagation
    sigma_qz = np.sqrt((np.sin(theta) * sigma_radius) ** 2 + (radius * np.cos(theta) * sigma_theta) ** 2)
    sigma_qxy = np.sqrt((np.cos(theta) * sigma_radius) ** 2 + (radius * np.sin(theta) * sigma_theta) ** 2)
    sigma_array = np.array([sigma_qz, sigma_qxy])

    qz = np.minimum(qz, qz_max)
    qxy = np.sqrt(radius ** 2 - qz ** 2)
    qxy = np.minimum(qxy, qxy_max)
    qz = np.sqrt(radius ** 2 - qxy ** 2)

    angle_deg = np.rad2deg(np.arctan2(qz, qxy))

    q_array = np.array([qz, qxy])
    return q_array, sigma_array, radius, angle_deg

def _data2container(boxes, polar_shape, q_abs_max, ang_deg_max, q_xy_max , q_z_max, wavelength):

    img_container = ImageContainer()
    img_container.amplitude = np.array([box.fitting_result['amplitude'] for box in boxes])
    img_container.A = np.array([box.fitting_result['A'] for box in boxes]) * polar_shape[1] / q_abs_max
    img_container.B = np.array([box.fitting_result['B'] for box in boxes]) * polar_shape[0] / ang_deg_max
    img_container.C = np.array([box.fitting_result['C'] for box in boxes])
    img_container.theta = np.array([box.fitting_result['theta'] for box in boxes])

    radius_arr = np.array([box.fitting_result['radius'] for box in boxes])/polar_shape[1] * q_abs_max
    angle_arr = np.array([box.fitting_result['angle'] for box in boxes]) / polar_shape[0] * ang_deg_max
    angle_arr = np.nan_to_num(angle_arr, nan=45)
    angle_arr[angle_arr < 2] = 0


    def adjust_missing_wedge(wavelength, q_abs, phi):
        k = 2 * np.pi / float(wavelength)

        q_abs = np.asarray(q_abs)
        phi = np.asarray(phi)

        mask = np.abs(q_abs) > np.abs(2 * k * np.cos(np.deg2rad(phi)))
        phi_corrected = phi.copy()

        phi_corrected[mask] = np.rad2deg(
            np.arccos(np.clip(q_abs[mask] / (2 * k), -1.0, 1.0))
        )

        return phi_corrected

    angle_arr = adjust_missing_wedge(wavelength, radius_arr, angle_arr)
    img_container.radius = radius_arr
    img_container.angle = angle_arr

    # scaling and *2 as in mlgidGUI and mlgidDETECT
    img_container.radius_width = np.array([box.fitting_result['radius_width'] for box in boxes])/polar_shape[1]*q_abs_max*2
    img_container.angle_width = np.array([box.fitting_result['angle_width'] for box in boxes])/polar_shape[0]*ang_deg_max*2

    img_container.amplitude_err = np.array([box.fitting_error['amplitude'] for box in boxes])
    img_container.A_err = np.array([box.fitting_error['A'] for box in boxes]) * polar_shape[1] / q_abs_max
    img_container.B_err = np.array([box.fitting_error['B'] for box in boxes]) * polar_shape[0] / ang_deg_max
    img_container.C_err = np.array([box.fitting_error['C'] for box in boxes])
    img_container.theta_err = np.array([box.fitting_error['theta'] for box in boxes])
    img_container.radius_err = np.array([box.fitting_error['radius'] for box in boxes])/polar_shape[1]*q_abs_max
    img_container.radius_width_err = np.array([box.fitting_error['radius_width'] for box in boxes])/polar_shape[1]*q_abs_max
    img_container.angle_err = np.array([box.fitting_error['angle'] for box in boxes])/polar_shape[0]*ang_deg_max
    img_container.angle_width_err = np.array([box.fitting_error['angle_width'] for box in boxes])/polar_shape[0]*ang_deg_max

    img_container.qzqxyboxes, img_container.qzqxyboxes_err, img_container.radius, img_container.angle = compute_qzqxy_with_error(
        img_container.radius, img_container.radius_err, img_container.angle, img_container.angle_err,
        q_z_max, q_xy_max)
    # img_container.qzqxyboxes = np.array([img_container.radius * np.sin(np.deg2rad(img_container.angle)),
    #                                      img_container.radius * np.cos(np.deg2rad(img_container.angle))])

    img_container.id = np.array([box.index for box in boxes])
    img_container.is_ring = np.array([box.is_ring for box in boxes])
    img_container.is_cut_qz = np.array([box.is_cut_qz for box in boxes])
    img_container.is_cut_qxy = np.array([box.is_cut_qxy for box in boxes])
    # img_container.q_xy = q_xy
    # img_container.q_z = q_z
    # img_container.visibility = visibility
    # img_container.score = score
    return img_container



class ProcessDataFromFile:
    """
    Process data from a NeXus file with clustering and Gaussian fitting.

    This class reads a NeXus file (converted images with detected boxes),
    processes a specific entry and frame (or all if None), performs clustering
    of peaks and rings, and fits the clusters with a set of one- or
    two-dimensional Gaussian functions.

    Parameters
    ----------
    filename : str
        Path to the NeXus file containing the data.
    entry : str, optional
        Specific entry in the file to process. If None, all entries are processed.
    frame_num : int, optional
        Frame number to process. If None, all frames are processed.
    crit_angle : float, optional
        Critical angle to shift the sample horizon (in degrees). Default is 0.0.
    polar_shape : Any, optional
        Shape of the polar array used for analysis. Default is [512, 1024].
    clustering_distance_peaks : float, optional
        Maximum distance (in pixels) for clustering peaks. Default is 10.0.
    clustering_distance_rings : float, optional
        Maximum distance (in pixels) for clustering rings. Default is 10.0.
    clustering_extend : float, optional
        Number of pixels to extend the cluster size. Default is 2.0.
    use_pool : bool, optional
        Whether to use peak pool from the previous image. Default is False.
    debug : bool, optional
        Enable debug mode to plot fitting results and parameters. Default is False.
    theta_fixed : bool, optional
        Whether to fix Gaussian tilt angle to 0° (azimuthal direction) during fitting. Default is True
    """

    def __init__(
        self,
        filename: str,
        entry: Optional[str] = None,
        frame_num: Optional[int] = None,
        crit_angle: float = 0.0,
        polar_shape: Any = None,
        clustering_distance_peaks: float = 10.0,
        clustering_distance_rings: float = 10.0,
        clustering_extend: float = 2.0,
        use_pool: bool = False,
        debug: bool = False,
        multiprocessing: bool = False,
        theta_fixed: bool = True
    ):
        self.filename = filename
        self.entry = entry
        self.frame_num = frame_num
        self.crit_angle = crit_angle
        self.polar_shape = polar_shape if polar_shape is not None else [512, 1024]
        self.clustering_distance_peaks = clustering_distance_peaks
        self.clustering_distance_rings = clustering_distance_rings
        self.clustering_extend = clustering_extend
        self.use_pool = use_pool
        self.debug = debug
        self.multiprocessing = multiprocessing
        self.theta_fixed = theta_fixed

        self.process_data_from_file()

    def process_data_from_file(self):
        """
        Process all entries or a specific entry from the NeXus file.

        This method opens the NeXus file, retrieves available entries, and
        processes either the specified entry or all entries sequentially.
        """
        self.nexus = pygid.NexusFile(self.filename)
        self.entry_dict = self.nexus.entry_dict
        if self.entry is None:
            for entry in self.entry_dict:
                self.process_single_entry(entry)
            return
        elif isinstance(self.entry, list):
            if not self.entry in self.entry_dict:
                raise ValueError("entry not found in the NeXus file")
            for entry in self.entry:
                self.process_single_entry(entry)
        else:
            if not self.entry in self.entry_dict:
                raise ValueError("entry not found in the NeXus file")
            self.process_single_entry(self.entry)

    def process_single_entry(self, entry):
        """
        Process a single entry from the NeXus file.

        Initializes internal state and processes either all frames or a specific frame.

        Parameters
        ----------
        entry : str
            Name of the entry to process.
        """
        self.yy = None
        self.zz = None
        self.ang_deg_max = None
        self.peaks_pool = [] if self.use_pool else None

        frame_num_all = self.entry_dict[entry]['shape'][0]

        if self.frame_num is None:
            for frame_num in range(frame_num_all):
                self.process_single_frame(entry, frame_num)
            return
        elif isinstance(self.frame_num, list):
            for frame_num in self.frame_num:
                if frame_num >= frame_num_all:
                    raise ValueError("frame_num is out of range")
                self.process_single_frame(entry, frame_num)
        else:
            if self.frame_num >= frame_num_all:
                raise ValueError("frame_num is out of range")
            self.process_single_frame(entry, self.frame_num)


    def process_single_frame(self, entry, frame_num):
        """
        Process a single frame of an entry: preprocess the image and detected peaks,
        perform clustering, and fit Gaussian functions.

        Parameters
        ----------
        entry : str
            Name of the NeXus entry to process.
        frame_num : int
            Index of the frame within the entry to process.

        Notes
        -----
        - Loads the converted image and parameters from the NeXus file.
        - Applies preprocessing.
        - Converts the image to polar coordinates.
        - Loads previously detected peaks from the NeXus file.
        - Performs Gaussian fitting and clustering on peaks and rings.
        - Updates peak pools if use_pool is enabled.
        - Saves the fit results back into the NeXus file.
        """
        # load conversion
        conversion = self.nexus.load_entry(entry, frame_num)
        q_xy = conversion.matrix[0].q_xy
        q_z = conversion.matrix[0].q_z
        img_reciprocal = conversion.img_gid_q[0]
        ai = conversion.params.ai[0]
        wavelength = conversion.params.wavelength

        # image preprocessing
        img_reciprocal = img_preprocessing(img_reciprocal, ai, self.crit_angle, wavelength, q_z)

        if self.yy is None or self.zz is None or self.ang_deg_max is None:
            self.yy, self.zz, self.ang_deg_max = (
                _get_polar_grid(img_reciprocal.shape, self.polar_shape, [0, 0]))
        polar_img = polar_conversion(img_reciprocal, self.yy, self.zz, cv2.INTER_LINEAR)

        # load detected peaks
        detected_peaks = read_detected_peaks(self.nexus, entry, frame_num)

        # run fitting, update pool
        img_container_fit, self.peaks_pool = fit_data(polar_img, detected_peaks['radius'],  detected_peaks['radius_width'],
                                     detected_peaks['angle'],  detected_peaks['angle_width'],
                                     wavelength, np.nanmax(q_xy), np.nanmax(q_z), np.sqrt(np.nanmax(q_z)**2+np.nanmax(q_xy)**2),
                                     self.ang_deg_max,
                                     self.clustering_distance_peaks,
                                     self.clustering_distance_rings, self.clustering_extend, self.theta_fixed,
                                     self.debug, self.multiprocessing, self.peaks_pool)

        img_container_fit.score = detected_peaks['score']
        img_container_fit.visibility = detected_peaks['visibility']
        img_container_fit.q_xy = q_xy
        img_container_fit.q_z = q_z

        img_container_fit.metadata = _set_fitting_metadata(
            clustering_distance_peaks=self.clustering_distance_peaks,
            clustering_distance_rings=self.clustering_distance_rings,
            clustering_extend=self.clustering_extend,
            use_pool=self.use_pool)

        # save img_container_fit
        save_fit(self.filename, entry, img_container_fit, frame_num)

def _set_fitting_metadata(**kwargs):
    metadata = {'program': 'pygidfit',
                'version': importlib.metadata.version("pygidfit"),
                'date': datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f'),
                }
    metadata.update(kwargs)
    return metadata


def fit_data(polar_img, radius, radius_width, angle, angle_width, wavelength, q_xy_max, q_z_max, q_abs_max, ang_deg_max = 90,
             clustering_distance_peaks = 10,
             clustering_distance_rings = 10, clustering_extend = 2, theta_fixed = False,
             debug = False, multiprocessing = False, peaks_pool = None):
    """
    Fit detected peaks in a polar image with Gaussian functions and cluster them.

    This function takes a polar-converted image and detected peaks (with their
    radii and angles), preprocesses the detected boxes, performs clustering, and applies
    Gaussian fitting to each cluster. Optionally, it updates a pool of peaks for
    sequential frame processing.

    Parameters
    ----------
    polar_img : ndarray
        Polar-converted 2D image to fit.
    radius : array-like
        Radii of the detected peaks.
    radius_width : array-like
        Widths of the detected peaks in radial direction.
    angle : array-like
        Angular positions of the detected peaks (in degrees).
    angle_width : array-like
        Angular widths of the detected peaks (in degrees).
    wavelength : float
        Wavelength used for the diffraction experiment.
    q_xy_max : float
        Maximum in-plane scattering vector.
    q_z_max : float
        Maximum out-of-plane scattering vector.
    q_abs_max : float
        Maximum absolute scattering vector.
    ang_deg_max : float, optional
        Maximum angle for polar image grid (default is 90).
    clustering_distance_peaks : float, optional
        Maximum distance (in pixels) for clustering peaks (default is 10).
    clustering_distance_rings : float, optional
        Maximum distance (in pixels) for clustering rings (default is 10).
    clustering_extend : float, optional
        Number of pixels to extend cluster size (default is 2).
    theta_fixed : bool, optional
        Whether to fix Gaussian tilt angle to 0° (azimuthal direction) during fitting. Default is True
    debug : bool, optional
        Enable debug mode with plots (default is False).
    multiprocessing : bool, optional
        Use multiprocessing for fitting (default is False).
    peaks_pool : list or None, optional
        Pool of peaks from previous frames; allows sequential fitting updates (default is None).

    Returns
    -------
    img_container : object
        Container object with fitted peak results, metadata, and image info.
    updated_peaks_pool : list or None
        Updated peaks pool if `peaks_pool` was provided, otherwise None.
    """
    polar_shape = polar_img.shape
    # boxes preprocessing
    detected_peaks = DetectedPeaks(radius = radius, radius_width = radius_width, angle = angle, angle_width = angle_width)
    boxes = boxes_preprocessing(detected_peaks,
                        polar_shape, wavelength,
                        q_abs_max,
                        q_xy_max, q_z_max)
    # clustarization
    clusters = cluster_boxes_by_centers(boxes, clustering_distance_peaks, clustering_distance_rings, clustering_extend)

    # real calling of fitting
    fit_single_image(polar_img, boxes, clusters, theta_fixed=theta_fixed, peaks_pool=peaks_pool, debug=debug,
                     multiprocessing=multiprocessing)

    img_container = _data2container(boxes, polar_shape, q_abs_max, ang_deg_max,
                                    q_xy_max, q_z_max,
                                    wavelength)
    if peaks_pool is None:
        return img_container, None
    else:
        return img_container, boxes




from dataclasses import dataclass, field
from typing import Tuple
import numpy as np


DEFAULT_POLAR_SHAPE: Tuple[int, int] = (512, 512)

@dataclass
class Labels():
    """Container for the labeled peaks. Exists for reciprocal- and polar shape."""

    boxes: list = field(default_factory=list)
    split_boxes: list = field(default_factory=list)
    radii: list = field(default_factory=list)
    widths: list = field(default_factory=list)
    angles: list = field(default_factory=list)
    angles_std: list = field(default_factory=list)
    confidences: list = field(default_factory=list)
    intensities: list = field(default_factory=list)
    background_levels: list = field(default_factory=list)
    background_slopes: list = field(default_factory=list)
    is_ring: list = field(default_factory=list)
    img_nr: int = 0
    img_name: str = None
    peak_height: int = 0

    
    def get_num_low_conf(self):
        return np.count_nonzero(self.confidences == 0.1)
    
    def get_num_med_conf(self):
        return np.count_nonzero(self.confidences == 0.5)
    
    def get_num_high_conf(self):
        return np.count_nonzero(self.confidences == 1)

    def __len__(self):
        return len(self.boxes)

@dataclass
class ImageContainer:
    """Container for GIWAXS images. Contains the image in raw and polar coordinates."""
    config = None
    raw_reciprocal: np.ndarray = None
    raw_polar_image: np.ndarray = None #field(init=False)
    converted_polar_image: np.ndarray = None #field(init=False)
    split_polar_images: np.array = None # field(init=False)
    #image mask, necessary to calculate the linear profile
    converted_mask: np.ndarray =  None #field(init=False)
    polar_img_shape: tuple = DEFAULT_POLAR_SHAPE
    q_z: float = None
    q_xy: float = None
    beam_center_x = None
    beam_center_y = None
    h5_group = None
    nr: int = None


    qzqxyboxes : np.array = None
    boxes: np.array = None
    scores: np.array = None
    radius: np.array = None
    radius_width: np.array = None
    angle: np.array = None
    angle_width = None
    reciprocal_labels: Labels = None # field(default_factory=Labels)
    polar_labels: Labels =  None #field(default_factory=Labels)

    def __post_init__(self):
        if self.config is not None:
            self.polar_img_shape: tuple = self.config.PREPROCESSING_POLAR_SHAPE



"""GIDFIT: Gaussian Fitting for Grazing Incidence Diffraction Data."""

from pygidfit.process_scans import ProcessDataFromFile, fit_data
# from pygidfit.main import run_scans, run_scans_img_container

__all__ = [
    "ProcessDataFromFile",
]
__version__ = "0.1.1"


import numpy as np
from lmfit import Model, Parameters
from lmfit.models import LinearModel
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.patches import Rectangle, Ellipse
from numba import jit
from multiprocessing import Pool, shared_memory
import time
from math import sin, cos, exp
import numexpr as ne

from pygidfit.box_utils import (
    make_box_attributes,
)

def safe_center_of_mass(arr):
    mask = np.isfinite(arr)
    if not np.any(mask):
        return np.nan, np.nan
    norm = np.nansum(arr)
    if norm == 0:
        return np.nan, np.nan
    y, x = np.indices(arr.shape)
    com_y = np.nansum(y * arr) / norm
    com_x = np.nansum(x * arr) / norm
    return com_y, com_x

# @jit(nopython=True, fastmath=True)
def two_d_rotated_gaussian(x, y, amp, xo, yo, sigma_x, sigma_y, theta):
    x0 = xo
    y0 = yo

    cos_t = cos(theta)
    sin_t = sin(theta)

    sin2 = 2.0 * sin_t * cos_t

    a = (cos_t * cos_t) / (2.0 * sigma_x * sigma_x) + (sin_t * sin_t) / (2.0 * sigma_y * sigma_y)
    b = -sin2 / (4.0 * sigma_x * sigma_x) + sin2 / (4.0 * sigma_y * sigma_y)
    c = (sin_t * sin_t) / (2.0 * sigma_x * sigma_x) + (cos_t * cos_t) / (2.0 * sigma_y * sigma_y)
    dx = x - x0
    dy = y - y0
    # out = np.empty_like(x)
    # for i in range(x.size):
    #     val = a * dx.flat[i] * dx.flat[i] + 2.0 * b * dx.flat[i] * dy.flat[i] + c * dy.flat[i] * dy.flat[i]
    #     out.flat[i] = amp * exp(-val)
    out = ne.evaluate("amp * exp(-(a*dx*dx + 2*b*dx*dy + c*dy*dy))")
    return out

@jit(nopython=True, fastmath=True)
def sum_of_gaussians_and_plane(x, y, param_array, n):
    z = np.zeros_like(x)
    for i in range(n):
        base = i * 6
        amp = param_array[base]
        xo = param_array[base + 1]
        yo = param_array[base + 2]
        sigx = param_array[base + 3]
        sigy = param_array[base + 4]
        theta = param_array[base + 5]

        cos_t = cos(theta)
        sin_t = sin(theta)
        sin2 = 2.0 * sin_t * cos_t

        a = (cos_t * cos_t) / (2.0 * sigx * sigx) + (sin_t * sin_t) / (2.0 * sigy * sigy)
        b = -sin2 / (4.0 * sigx * sigx) + sin2 / (4.0 * sigy * sigy)
        c = (sin_t * sin_t) / (2.0 * sigx * sigx) + (cos_t * cos_t) / (2.0 * sigy * sigy)

        for k in range(x.size):
            dx = x.flat[k] - xo
            dy = y.flat[k] - yo
            z.flat[k] += amp * exp(-(a*dx*dx + 2*b*dx*dy + c*dy*dy))

    idx_plane = n * 6
    a = param_array[idx_plane] if idx_plane < param_array.size else 0.0
    b = param_array[idx_plane + 1] if idx_plane + 1 < param_array.size else 0.0
    c = param_array[idx_plane + 2] if idx_plane + 2 < param_array.size else 0.0

    for k in range(x.size):
        z.flat[k] += a * x.flat[k] + b * y.flat[k] + c

    return z

def build_sum_gaussians_wrapper(n):
    def model_func(x, y, **params):
        param_array = np.empty(n*6 + 3, dtype=np.float64)
        pos = 0

        for i in range(n):
            param_array[pos]   = params[f'g{i}_amplitude']
            param_array[pos+1] = params[f'g{i}_radius']
            param_array[pos+2] = params[f'g{i}_angle']
            param_array[pos+3] = params[f'g{i}_radius_width']
            param_array[pos+4] = params[f'g{i}_angle_width']
            param_array[pos+5] = params[f'g{i}_theta']
            pos += 6

        param_array[pos]   = params['A']
        param_array[pos+1] = params['B']
        param_array[pos+2] = params['C']

        return sum_of_gaussians_and_plane(x, y, param_array, n)
    return model_func

def compute_initial_params(sub, x0, y0, x1, y1, debug = False):
    """Compute initial 2D Gaussian parameters from subarray."""
    amp = np.nanmax(sub)
    if debug:
        print("compute_initial_params")
    com_y, com_x = safe_center_of_mass(sub-np.nanpercentile(sub, 10))
    xo = x0 + com_x
    yo = y0 + com_y
    sigma_x = max((x1 - x0) / 2 , 1.0)
    sigma_y = max((y1 - y0) / 2 , 1.0)
    return amp, xo, yo, sigma_x, sigma_y


def fit_peak_cluster(cluster, boxes, img, peaks_pool = None, theta_fixed=False, debug=False):
    """Fit a cluster of 2D Gaussian peaks with a background plane over the bounding box."""
    # Extract ROI bounding box from the cluster
    time0 = time.time()

    xmin, ymin, xmax, ymax = np.round(cluster.bbox).astype(int)

    h, w = img.shape
    xmin = np.clip(xmin, 0, w)
    xmax = np.clip(xmax, 0, w)
    ymin = np.clip(ymin, 0, h)
    ymax = np.clip(ymax, 0, h)


    # Extract ROI from the image
    roi = img[ymin:ymax, xmin:xmax]
    mask = np.isfinite(roi)
    for (mxmin, mymin, mxmax, mymax) in cluster.mask_boxes:
        rxmin = int(np.clip(mxmin - xmin, 0, roi.shape[1]))
        rxmax = int(np.clip(mxmax - xmin, 0, roi.shape[1]))
        rymin = int(np.clip(mymin - ymin, 0, roi.shape[0]))
        rymax = int(np.clip(mymax - ymin, 0, roi.shape[0]))
        mask[rymin:rymax, rxmin:rxmax] = False
    roi[~mask] = np.nan

    Y, X = np.mgrid[ymin:ymax, xmin:xmax]
    data = roi.ravel()
    X_flat = X.ravel()
    Y_flat = Y.ravel()

    valid_mask = np.isfinite(data)
    data = data[valid_mask]
    X_flat = X_flat[valid_mask]
    Y_flat = Y_flat[valid_mask]

    # Plane params
    a_bkg, b_bkg, c_bkg = 0, 0, np.nanpercentile(roi, 10)
    X_plane, Y_plane = X, Y
    Z_plane = a_bkg * X_plane + b_bkg * Y_plane + c_bkg
    roi_corrected = roi - Z_plane

    # Build sum of 2D Gaussians model
    # model = Model(build_sum_gaussians(len(cluster.indices)), independent_vars=['x', 'y'])
    model = Model(build_sum_gaussians_wrapper(len(cluster.indices)), independent_vars=['x', 'y'])

    params = Parameters()

    # Loop through all boxes in the cluster
    for i, idx in enumerate(cluster.indices):
        box = boxes[idx]
        x0 = int(np.clip(np.round(box.limits[0]), xmin, xmax ))
        y0 = int(np.clip(np.round(box.limits[1]) ,ymin, ymax))
        x1 = int(np.clip(np.round(box.limits[2]) , xmin, xmax))
        y1 = int(np.clip(np.round(box.limits[3]) , ymin, ymax))

        sub = roi_corrected[y0- ymin:y1- ymin, x0- xmin:x1- xmin]

        # Skip empty or invalid boxes
        if sub.size == 0 or not np.isfinite(sub).any():
            if debug:
                print("invalid")
            amp = 0
            xo = (x0 + x1) / 2
            yo = (y0 + y1) / 2
            sigma_x = max((x1 - x0) / 2 , 1.0)
            sigma_y = max((y1 - y0) / 2 , 1.0)
        else:
            prev_box = None
            if peaks_pool is not None and len(peaks_pool) > 0:
                for b in peaks_pool:
                    r, a = b.fitting_result['radius'], b.fitting_result['angle']
                    # print("box.limits[0], r, box.limits[2],  box.limits[1] , a, box.limits[3] ",box.limits[0], r, box.limits[2],  box.limits[1] , a, box.limits[3])
                    if box.limits[0] <= r <= box.limits[2] and box.limits[1] <= a <= box.limits[3]:
                        prev_box = b
                        break
            if prev_box is not None:
                if debug:
                    print("Use previous peak")
                amp = prev_box.fitting_result['amplitude']
                xo = prev_box.fitting_result['radius']
                yo = prev_box.fitting_result['angle']
                sigma_x = prev_box.fitting_result['radius_width']
                sigma_y = prev_box.fitting_result['angle_width']
            else:
                if debug and peaks_pool is not None:
                    print("Couldn't find previous peak")
                amp, xo, yo, sigma_x, sigma_y = compute_initial_params(sub, x0, y0, x1, y1, debug)
        # go to the sample horizon
        if y0 < h/90*5:
            y0 = 0
            vary_y0 = False
        else:
            vary_y0 = True
        if debug:
            print("amp, xo, yo, sigma_x, sigma_y, vary_y0 ", amp, xo, yo, sigma_x, sigma_y, vary_y0)
        x_bound_min = np.clip(x0 - (x1 - x0)/4, xmin, xmax)
        x_bound_max = np.clip(x1 + (x1 - x0)/4 , xmin, xmax)
        y_bound_min = np.clip(y0 - (y1 - y0)/4 , ymin, ymax)
        y_bound_max = np.clip(y1 + (y1 - y0)/4 , ymin, ymax) if not box.is_cut_qz else h

        # if x_bound_min == x_bound_max:
        #     print(x_bound_min, x_bound_max, x0, x1, xmin, xmax)
        #     print(box.limits)
        # Add Gaussian parameters to the model
        params.add(f'g{i}_amplitude', value=amp, min=0)
        params.add(f'g{i}_radius', value=xo, min=x_bound_min, max=x_bound_max)
        params.add(f'g{i}_angle', value=yo, min=y_bound_min, max=y_bound_max, vary = vary_y0)
        params.add(f'g{i}_radius_width', value=sigma_x, min=(x1-x0)/8, max = (x1-x0)/2)
        params.add(f'g{i}_angle_width', value=sigma_y, min=(y1-y0)/8, max = (y1-y0)/2)
        params.add(f'g{i}_theta', value=0, vary=not theta_fixed)

    # Add parameters for the background plane
    params.add('A', value=a_bkg, min = -0.1, max = 0.1)
    params.add('B', value=b_bkg, min = -1, max = 1)
    params.add('C', value=c_bkg, min = -abs(c_bkg/4)-1, max = abs(c_bkg*2)+1)

    # Exit if no valid peaks
    if len(params) == 0:
        return None, None, None
    time1 = time.time()

    if False:
        res = fit_peak_cluster_jaxfit_from_lmfit(data, X_flat, Y_flat, params)
    if False:
        res = fit_peak_cluster_numba_cuda_from_lmfit(data, X_flat, Y_flat, params, max_nfev=500)

    # Perform the fit
    try:
        result = model.fit(data, params=params, x=X_flat, y=Y_flat, max_nfev=500, method="least_squares") # max_nfev=100
        list_to_return = {
            'params': dict(result.best_values),
            'errors': {
                name: (result.params[name].stderr if result.params[name].stderr is not None else np.nan)
                for name in result.params
            },
            'success': result.success,
            'message': result.message,
        }

    except:
        # print("Failed")
        param_values = {name: p.value for name, p in params.items()}
        param_errors = {name: np.nan for name in params}
        result = None
        list_to_return = {
            'params': param_values,
            'errors': param_errors,
            'success': False,
            'message': 'fit failed'
        }
    time2 = time.time()

    if debug:
        if params is not None:
            plot_peak_cluster_debug(
                roi=roi,
                xmin=xmin,
                ymin=ymin,
                cluster=cluster,
                boxes=boxes,
                params=params,
                result=result,
                time_preproc=(time1 - time0),
                time_fit=(time2 - time1)
            )
        else:
            print("Parameters not available")

    return list_to_return


def plot_peak_cluster_debug(roi, xmin, ymin, cluster, boxes, params, result, time_preproc, time_fit):
    """
    Visualizes a ROI with bounding boxes and ellipses for initial guesses and fitted Gaussians.

    Parameters
    ----------
    roi : np.ndarray
        Region of Interest (image section).
    xmin, ymin : float
        ROI offset relative to the original image.
    cluster : object
        Cluster object containing `indices`.
    boxes : list
        List of box objects (with `.limits` attribute).
    params : lmfit.Parameters
        Initial Gaussian parameters.
    result : lmfit.ModelResult
        Fit results.
    time_preproc, time_fit : float
        Preprocessing and fitting times in seconds.
    """
    if not np.any(roi > 0):
        return
    norm = LogNorm(vmin=np.nanmin(roi[roi > 0]), vmax=np.nanmax(roi))
    fig, axes = plt.subplots(figsize=(6, 6))
    xmin, ymin, xmax, ymax = np.round(cluster.bbox).astype(int)
    axes.imshow(roi, cmap='inferno', origin='lower', norm=norm, extent=[xmin, xmax, ymin, ymax])

    # Draw boxes
    for i in cluster.indices:
        box = boxes[i]
        x = box.limits[0] #- xmin
        y = box.limits[1] #- ymin
        w = box.limits[2] - box.limits[0]
        h = box.limits[3] - box.limits[1]
        rect = Rectangle((x, y), w, h, linewidth=5, edgecolor='red',
                         facecolor='None', alpha=1)
        axes.add_patch(rect)
    axes.set_title(str(cluster.indices))

    # Initial peaks
    for i, idx in enumerate(cluster.indices):
        amp = params.get(f'g{i}_amplitude', None)
        xo = params.get(f'g{i}_radius', None)
        yo = params.get(f'g{i}_angle', None)
        sigma_x = params.get(f'g{i}_radius_width', None)
        sigma_y = params.get(f'g{i}_angle_width', None)

        if None in [amp, xo, yo, sigma_x, sigma_y]:
            continue

        ellipse = Ellipse(
            (xo.value, yo.value),
            width=2 * sigma_x.value,
            height=2 * sigma_y.value,
            edgecolor='blue',
            facecolor='none',
            linewidth=2,
            alpha=1,
            linestyle='--',
            label=f'init peak {idx}'
        )
        theta = params.get(f'g{i}_theta', None)
        if theta is not None:
            ellipse.angle = np.degrees(-theta.value)
        axes.add_patch(ellipse)

    # Fitted peaks
    for i in range(len(cluster.indices)):
        try:
            amp = result.params[f'g{i}_amplitude']
            xo = result.params[f'g{i}_radius']
            yo = result.params[f'g{i}_angle']
            sigma_x = result.params[f'g{i}_radius_width']
            sigma_y = result.params[f'g{i}_angle_width']
            theta = result.params[f'g{i}_theta']
        except KeyError:
            continue

        ellipse_fit = Ellipse(
            (xo.value, yo.value),
            width=2 * sigma_x.value,
            height=2 * sigma_y.value,
            angle=np.degrees(-theta.value),
            edgecolor='green',
            facecolor='none',
            linewidth=3,
            alpha=1,
            linestyle='--',
            label=f'fit peak {cluster.indices[i]}'
        )
        axes.add_patch(ellipse_fit)

    axes.set_aspect('auto')
    axes.legend()
    plt.show()

    print(f"Preprocessing took {time_preproc * 1000:.2f} ms")
    print(f"Fitting took {time_fit * 1000:.2f} ms")

def sum_of_gaussians_and_plane_and_1d(x, y, param_array, n, m):
    z = np.zeros_like(x, dtype=np.float64)

    # 2D rotated Gaussians
    for i in range(n):
        base = i * 6
        amp = param_array[base]
        xo = param_array[base + 1]
        yo = param_array[base + 2]
        sigx = param_array[base + 3]
        sigy = param_array[base + 4]
        theta = param_array[base + 5]

        cos_t = cos(theta)
        sin_t = sin(theta)
        sin2 = 2.0 * sin_t * cos_t

        a = (cos_t * cos_t) / (2.0 * sigx * sigx) + (sin_t * sin_t) / (2.0 * sigy * sigy)
        b = -sin2 / (4.0 * sigx * sigx) + sin2 / (4.0 * sigy * sigy)
        c = (sin_t * sin_t) / (2.0 * sigx * sigx) + (cos_t * cos_t) / (2.0 * sigy * sigy)

        dx = x - xo
        dy = y - yo
        z += ne.evaluate("amp * exp(-(a*dx*dx + 2*b*dx*dy + c*dy*dy))")

    # Plane background
    offset_plane = n * 6
    a_plane = param_array[offset_plane]
    b_plane = param_array[offset_plane + 1]
    c_plane = param_array[offset_plane + 2]

    z += ne.evaluate("a_plane * x + b_plane * y + c_plane")

    # 1D Gaussians along x
    offset_1d = offset_plane + 3
    for j in range(m):
        base = offset_1d + j * 3
        amp_1d = param_array[base]
        center_1d = param_array[base + 1]
        sigma_1d = param_array[base + 2]

        dx = x - center_1d
        inv2sigma2 = 0.5 / (sigma_1d * sigma_1d)
        z += ne.evaluate("amp_1d * exp(-dx*dx*inv2sigma2)")

    return z

def gaussian_height(x, radius, amplitude, radius_width):
    dx = x - radius
    inv2sigma2 = 0.5 / (radius_width * radius_width)
    return ne.evaluate("amplitude * exp(-dx*dx*inv2sigma2)")


def build_sum_gaussians_and_1d_wrapper(n, m):
    def model_func(x, y, **params):
        param_list = []

        # 2D Gaussians
        for i in range(n):
            for key in ['amplitude', 'radius', 'angle', 'radius_width', 'angle_width', 'theta']:
                param_list.append(params[f'g{i}_{key}'])

        # Background plane (MUST come before 1D Gaussians)
        param_list.extend([params['A'], params['B'], params['C']])

        # 1D Gaussians (only X-dependent)
        for i in range(m):
            for key in ['amplitude', 'radius', 'radius_width']:
                param_list.append(params[f'g1d_{i}_{key}'])

        param_array = np.array(param_list, dtype=np.float64)
        return sum_of_gaussians_and_plane_and_1d(x, y, param_array, n, m)

    return model_func

def visualize_fit_3d(X, Y, Z_data, Z_fit):

    fig = plt.figure(figsize=(10, 6))
    ax = fig.add_subplot(111, projection='3d')

    ax.plot_surface(X, Y, Z_data, cmap='inferno', alpha=0.5, rstride=1, cstride=1, edgecolor='none')
    ax.plot_surface(X, Y, Z_fit, cmap='inferno', alpha=0.5, rstride=1, cstride=1, edgecolor='none')

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Intensity')
    ax.set_title('3D Fit vs Data')
    plt.tight_layout()
    plt.show()



def fit_peak_on_ring_cluster(cluster, boxes, img, peaks_pool, theta_fixed, debug = False):
    time0 = time.time()
    xmin, ymin, xmax, ymax = np.round(cluster.bbox).astype(int)
    h, w = img.shape
    xmin = np.clip(xmin, 0, w)
    xmax = np.clip(xmax, 0, w)
    ymin = np.clip(ymin, 0, h)
    ymax = np.clip(ymax, 0, h)

    # Extract ROI from the image
    roi = img[ymin:ymax, xmin:xmax]
    y_len, x_len = roi.shape

    # Create grid coordinates
    # Y, X = np.mgrid[0:y_len, 0:x_len]
    Y, X = np.mgrid[ymin:ymax, xmin:xmax]
    data = roi.ravel()
    X_flat = X.ravel()
    Y_flat = Y.ravel()

    valid_mask = np.isfinite(data)
    data = data[valid_mask]
    X_flat = X_flat[valid_mask]
    Y_flat = Y_flat[valid_mask]

    # Plane params
    a, b, c = 0, 0, np.nanpercentile(roi, 10)
    X_plane, Y_plane = np.mgrid[ymin: ymax, xmin: xmax]
    Z_plane = a * X_plane + b * Y_plane + c
    roi_corrected = roi - Z_plane

    # Build sum of 2D Gaussians model

    peak_indices = []
    ring_indices = []
    for i in cluster.indices:
        if boxes[i].is_ring:
            ring_indices.append(i)
        else:
            peak_indices.append(i)

    model = Model(build_sum_gaussians_and_1d_wrapper(len(peak_indices), len(ring_indices)), independent_vars=['x', 'y'])
    params = Parameters()


    # Loop through all boxes in the cluster
    for i, idx in enumerate(peak_indices):
        box = boxes[idx]
        x0 = int(np.clip(np.round(box.limits[0]), xmin, xmax ))
        y0 = int(np.clip(np.round(box.limits[1]) ,ymin, ymax))
        x1 = int(np.clip(np.round(box.limits[2]) , xmin, xmax))
        y1 = int(np.clip(np.round(box.limits[3]) , ymin, ymax))

        # sub = roi_corrected[y0:y1, x0:x1]
        sub = roi_corrected[y0 - ymin:y1 - ymin, x0 - xmin:x1 - xmin]

        if sub.size == 0 or not np.isfinite(sub).any():
            # continue
            amp = 0
            xo = (x0 + x1) / 2
            yo = (y0 + y1) / 2
            sigma_x = max((x1 - x0) / 2 / 2.355, 1.0)
            sigma_y = max((y1 - y0) / 2 / 2.355, 1.0)
        else:
        # Estimate initial parameters
            amp = np.nanmax(sub)
            yy, xx = np.indices(sub.shape)
            com_y, com_x = safe_center_of_mass(sub)
            xo = x0 + com_x
            yo = y0 + com_y
            sigma_x = max((x1 - x0) /2/2.355, 1.0)  # FWHM to sigma conversion
            sigma_y = max((y1 - y0) /2/2.355, 1.0)

        # go to the sample horizon
        if y0 < h/90*5:
            y0 = 0
            vary_y0 = False
        else:
            vary_y0 = True
        if debug:
            print("amp, xo, yo, sigma_x, sigma_y, vary_y0 ", amp, xo, yo, sigma_x, sigma_y, vary_y0)

        x_bound_min = np.clip(x0 - (x1 - x0)/4, xmin, xmax)
        x_bound_max = np.clip(x1 + (x1 - x0)/4 , xmin, xmax)
        y_bound_min = np.clip(y0 - (y1 - y0)/4 , ymin, ymax)
        y_bound_max = np.clip(y1 + (y1 - y0)/4 , ymin, ymax) if not box.is_cut_qz else h

        # Add Gaussian parameters to the model
        params.add(f'g{i}_amplitude', value=amp, min=0)
        params.add(f'g{i}_radius', value=xo, min=x_bound_min, max=x_bound_max)
        params.add(f'g{i}_angle', value=yo, min=y_bound_min, max=y_bound_max, vary = vary_y0)
        params.add(f'g{i}_radius_width', value=sigma_x, min=(x1 - x0) / 8, max=(x1 - x0) / 2)
        params.add(f'g{i}_angle_width', value=sigma_y, min=(y1 - y0) / 8, max=(y1 - y0) / 2)
        params.add(f'g{i}_theta', value=0, vary=not theta_fixed)

    # params.add('A', value=a, min = -1, max = 1)
    params.add('A', value=a, min=-0.1, max=0.1)
    params.add('B', value=b, min = -1, max = 1)
    params.add('C', value=c, min = -abs(c/4)-1, max = abs(c*2)+1)

    for j, idx in enumerate(ring_indices):
        box = boxes[idx]

        center = box.fitting_result['radius']
        sigma = box.fitting_result['radius_width']
        amp = box.fitting_result['amplitude']

        params.add(f'g1d_{j}_amplitude', value=amp, min=0)
        params.add(f'g1d_{j}_radius', value=center, vary=False)
        params.add(f'g1d_{j}_radius_width', value=sigma, vary=False)


    # Exit if no valid peaks
    if len(params) == 0:
        return None, None, None
    time1 = time.time()

    for name, p in params.items():
        if np.isnan(p.value):
            if "amplitude" in name:
                p.value = 1.0
            else:
                p.value = 0.0

    # Perform the fit
    result = model.fit(data, params=params, x=X_flat, y=Y_flat, max_nfev=500, method="least_squares")  # max_nfev=100
    time2 = time.time()

    if debug:
        plot_peak_on_ring_cluster_debug(
            X=X,
            Y=Y,
            roi=roi,
            X_flat=X_flat,
            Y_flat=Y_flat,
            xmin=xmin,
            ymin=ymin,
            model=model,
            result=result,
            peak_indices=peak_indices,
            ring_indices=ring_indices,
            cluster=cluster,
            boxes=boxes,
            params=params,
            time_preproc=(time1 - time0),
            time_fit=(time2 - time1),
            visualize_fit_3d_func=visualize_fit_3d
        )

    list_to_return = {
        'params': dict(result.best_values),
        'errors': {
            name: (result.params[name].stderr if result.params[name].stderr is not None else np.nan)
            for name in result.params
        },
        'success': result.success,
        'message': result.message,
    }

    return list_to_return



def plot_peak_on_ring_cluster_debug(X, Y, roi, X_flat, Y_flat, xmin, ymin,
                   model, result, peak_indices, ring_indices,
                   cluster, boxes, params, time_preproc, time_fit,
                   visualize_fit_3d_func):
    """
    Visualizes the ROI with bounding boxes for peaks and rings,
    along with initial guesses and fitted Gaussian ellipses.

    Parameters
    ----------
    X, Y : np.ndarray
        Meshgrid arrays for the ROI.
    roi : np.ndarray
        Region of Interest (image section).
    X_flat, Y_flat : np.ndarray
        Flattened coordinates for model evaluation.
    xmin, ymin : float
        ROI offset relative to the original image.
    model : lmfit.Model
        The fitted model.
    result : lmfit.ModelResult
        Fit results containing parameters.
    peak_indices : list[int]
        Indices of detected peaks.
    ring_indices : list[int]
        Indices of detected rings.
    cluster : object
        Cluster object containing `.indices`.
    boxes : list
        List of box objects (with `.limits` attribute).
    params : lmfit.Parameters
        Initial Gaussian parameters.
    time_preproc, time_fit : float
        Preprocessing and fitting times in seconds.
    visualize_fit_3d_func : callable
        Function for 3D visualization, signature: (X, Y, roi, Z_fit_full).
    """

    # Model evaluation
    Z_fit_valid = model.eval(params=result.params, x=X_flat, y=Y_flat)
    Z_fit_full = np.full_like(roi, np.nan, dtype=np.float64)
    Z_fit_full[np.isfinite(roi)] = Z_fit_valid

    # Optional 3D visualization
    visualize_fit_3d_func(X, Y, roi, Z_fit_full)

    fig, axes = plt.subplots(figsize=(6, 6))
    norm = LogNorm(vmin=np.nanmin(roi[roi > 0]), vmax=np.nanmax(roi))
    xmin, ymin, xmax, ymax = np.round(cluster.bbox).astype(int)
    axes.imshow(roi, cmap='inferno', origin='lower', norm=norm, extent=[xmin, xmax, ymin, ymax])

    # Peak bounding boxes
    for i in peak_indices:
        box = boxes[i]
        x = box.limits[0] # - xmin
        y = box.limits[1] #- ymin
        w = box.limits[2] - box.limits[0]
        h = box.limits[3] - box.limits[1]
        rect = Rectangle((x, y), w, h, linewidth=5, edgecolor='red',
                         facecolor='None', alpha=1)
        axes.add_patch(rect)

    # Ring bounding boxes
    for i in ring_indices:
        box = boxes[i]
        x = box.limits[0] # box.limits[0] - xmin
        y = ymin # box.limits[1] - ymin
        w = box.limits[2] - box.limits[0]
        h = ymax - ymin# box.limits[3] - box.limits[1]
        rect = Rectangle((x, y), w, h, linewidth=5, edgecolor='black',
                         facecolor='None', alpha=1, linestyle='--',
                         label=f'ring box {i}')
        axes.add_patch(rect)

    axes.set_title(str(cluster.indices))

    # Initial Gaussian ellipses
    for i, idx in enumerate(cluster.indices):
        amp = params.get(f'g{i}_amplitude', None)
        xo = params.get(f'g{i}_radius', None)
        yo = params.get(f'g{i}_angle', None)
        sigma_x = params.get(f'g{i}_radius_width', None)
        sigma_y = params.get(f'g{i}_angle_width', None)

        if None in [amp, xo, yo, sigma_x, sigma_y]:
            continue

        ellipse = Ellipse(
            (xo.value, yo.value),
            width=2 * sigma_x.value,
            height=2 * sigma_y.value,
            edgecolor='white',
            facecolor='none',
            linewidth=2,
            alpha=1,
            linestyle='--',
            label=f'init peak {idx}'
        )
        theta = params.get(f'g{i}_theta', None)
        if theta is not None:
            ellipse.angle = np.degrees(theta.value)

        axes.add_patch(ellipse)

    # Fitted Gaussian ellipses
    for i in range(len(peak_indices)):
        try:
            amp = result.params[f'g{i}_amplitude']
            xo = result.params[f'g{i}_radius']
            yo = result.params[f'g{i}_angle']
            sigma_x = result.params[f'g{i}_radius_width']
            sigma_y = result.params[f'g{i}_angle_width']
            theta = result.params[f'g{i}_theta']
        except KeyError:
            continue

        ellipse_fit = Ellipse(
            (xo.value, yo.value),
            width=2 * sigma_x.value,
            height=2 * sigma_y.value,
            angle=np.degrees(theta.value),
            edgecolor='green',
            facecolor='none',
            linewidth=3,
            alpha=1,
            linestyle='--',
            label=f'fit peak {cluster.indices[i]}'
        )
        axes.add_patch(ellipse_fit)

    axes.set_aspect('auto')
    axes.legend()
    plt.show()

    print(f"Preprocessing took {time_preproc * 1000:.2f} ms")
    print(f"Fitting took {time_fit * 1000:.2f} ms")

def fit_ring_cluster(cluster, boxes, img,  peaks_pool, debug = False):
    xmin, ymin, xmax, ymax = np.round(cluster.bbox).astype(int)
    h, w = img.shape


    xmin = np.clip(xmin, 0, w)
    xmax = np.clip(xmax, 0, w)
    ymin = np.clip(ymin, 0, h)
    ymax = np.clip(ymax, 0, h)

    roi = img[ymin:ymax, xmin:xmax]
    profile = np.nanmean(roi, axis=0)

    x = np.arange(xmin, xmax)

    # Estimate linear background
    mid = len(profile) // 2

    left = profile[:mid]
    right = profile[mid:]

    if np.all(np.isnan(left)) or np.all(np.isnan(right)):
        ind0 = 0
        ind1 = len(profile) - 1
    else:
        ind0 = np.nanargmin(left)
        ind1 = np.nanargmin(right) + mid

    x0_lin, x1_lin = x[ind0], x[ind1]
    y0_lin, y1_lin = profile[ind0], profile[ind1]
    slope_guess = (y1_lin - y0_lin) / (x1_lin - x0_lin)
    intercept_guess = y0_lin - slope_guess * x0_lin
    if np.isnan(slope_guess) or np.isnan(intercept_guess):
        slope_guess = 0.0
        intercept_guess = 0.0
    background = slope_guess * x + intercept_guess
    profile_corrected = profile - background

    # Background model
    model = LinearModel(prefix='lin_')
    params = model.make_params(intercept=intercept_guess, slope=slope_guess)

    # Add one Gaussian per ring
    for i, idx in enumerate(cluster.indices):
        box = boxes[idx].limits
        x0_box = int(np.round(box[0]))
        x1_box = int(np.round(box[2]))

        x0_box = np.clip(x0_box, 0, w)
        x1_box = np.clip(x1_box, 0, w)

        center_guess = (x0_box + x1_box) / 2
        x0_rel = max(x0_box - xmin, 0)
        x1_rel = min(x1_box - xmin, len(profile_corrected))
        if x1_rel <= x0_rel:
            height_guess = 0
            sigma_guess = 1
        else:
            height_guess = np.nanmax(profile_corrected[x0_rel:x1_rel])
            sigma_guess = (x1_box - x0_box) / 4

        gmod = Model(gaussian_height, prefix=f'g{i}_')
        gparams = gmod.make_params(
            radius=center_guess,
            amplitude=height_guess,
            radius_width=sigma_guess
        )
        gparams[f'g{i}_amplitude'].min = 0
        gparams[f'g{i}_radius_width'].min = 0

        params.update(gparams)
        model += gmod

    try:
        mask = np.isfinite(profile)
        profile = profile[mask]
        x =  x[mask]
        result = model.fit(profile, params, x=x, max_nfev=500, method="least_squares",) #, max_nfev=100
        param_values = dict(result.best_values)
        param_errors = {
            name: (result.params[name].stderr if result.params[name].stderr is not None else np.nan)
            for name in result.params
        }
        success = result.success
        message = result.message
    except:
        param_values = {name: p.value for name, p in params.items()}
        param_errors = {name: np.nan for name in params}
        success = False
        message = "Failed"

    if debug:
        plt.figure(figsize=(6, 4))
        plt.plot(x, profile, 'b', label='Data')
        try:
            plt.plot(x, result.best_fit, 'r-', label='Best Fit')
            plt.plot(x, result.init_fit, 'c--', label='Initial Guess')
        except:
            print("Fitting failed")
        plt.title(f"Cluster {cluster.indices.tolist()}")
        plt.xlabel('X [pixels]')
        plt.ylabel('Mean intensity')
        plt.legend()
        plt.tight_layout()
        plt.show()

    return {
        'params': param_values,
        'errors': param_errors,
        'success': success,
        'message': message,
    }



def process_cluster_args(args):
    cluster, cluster_type, boxes, img, masked_img, theta_fixed, debug = args
    if cluster_type == 'rings':
        result = fit_ring_cluster(cluster, boxes, masked_img, None, debug)
    elif cluster_type == 'peaks':
        result = fit_peak_cluster(cluster, boxes, img, None, theta_fixed, debug)
    elif cluster_type == 'both':
        result = fit_peak_on_ring_cluster(cluster, boxes, img, None, theta_fixed, debug)
    else:
        result = None
    return cluster, result

##### MP with shared memory
def fit_clusters_multiprocessing(clusters, boxes, img, masked_img, theta_fixed, debug=False):
    cluster_types = ['rings', 'peaks', 'both']

    shm_img, shm_masked, img_shape, masked_shape, img_dtype, masked_dtype = init_shared_images(img, masked_img)

    try:
        for ctype in cluster_types:
            ctype_clusters = [(cluster, ctype, boxes,
                               shm_img.name, shm_masked.name,
                               img_shape, masked_shape, img_dtype, masked_dtype, theta_fixed,
                               debug)
                              for cluster in clusters if cluster.type == ctype]
            if not ctype_clusters:
                continue

            with Pool() as pool:
                results = pool.map(process_cluster_shared, ctype_clusters)

            for cluster, fitting_result in results:
                make_box_attributes(cluster.indices, boxes, fitting_result, cluster.type, debug)
                cluster.fitting_result = fitting_result
    finally:
        shm_img.close()
        shm_img.unlink()
        shm_masked.close()
        shm_masked.unlink()


def init_shared_images(img, masked_img):
    shm_img = shared_memory.SharedMemory(create=True, size=img.nbytes)
    shm_masked = shared_memory.SharedMemory(create=True, size=masked_img.nbytes)

    shm_img_array = np.ndarray(img.shape, dtype=img.dtype, buffer=shm_img.buf)
    shm_masked_array = np.ndarray(masked_img.shape, dtype=masked_img.dtype, buffer=shm_masked.buf)
    np.copyto(shm_img_array, img)
    np.copyto(shm_masked_array, masked_img)

    return shm_img, shm_masked, img.shape, masked_img.shape, img.dtype, masked_img.dtype

def process_cluster_shared(args):
    cluster, ctype, boxes, shm_img_name, shm_masked_name, img_shape, masked_shape, img_dtype, masked_dtype, theta_fixed, debug = args

    existing_shm_img = shared_memory.SharedMemory(name=shm_img_name)
    existing_shm_masked = shared_memory.SharedMemory(name=shm_masked_name)
    img = np.ndarray(img_shape, dtype=img_dtype, buffer=existing_shm_img.buf)
    masked_img = np.ndarray(masked_shape, dtype=masked_dtype, buffer=existing_shm_masked.buf)

    fitting_result = process_cluster_args((cluster, ctype, boxes, img, masked_img, theta_fixed, debug))

    return fitting_result


############## numba start

# import numpy as np
# from numba import cuda
# from scipy.optimize import least_squares
# import cupy as cp
# import math
#
# @cuda.jit
# def compute_residuals_kernel(X, Y, zdata, params, residuals, N_gauss):
#     idx = cuda.grid(1)
#     if idx >= zdata.size:
#         return
#
#     val = 0.0
#     for g in range(N_gauss):
#         base = g*6
#         n0 = params[base]
#         x0 = params[base+1]
#         y0 = params[base+2]
#         sigma_x = params[base+3]
#         sigma_y = params[base+4]
#         theta = params[base+5]
#
#         Xr = X[idx] - x0
#         Yr = Y[idx] - y0
#         gauss = math.exp(-0.5*(Xr**2/(sigma_x*sigma_x) + Yr**2/(sigma_y*sigma_y))) * n0
#         val += gauss
#
#     A = params[-3]
#     B = params[-2]
#     C = params[-1]
#     val += A + B*X[idx] + C*Y[idx]
#
#     residuals[idx] = zdata[idx]
#     residuals[idx] -= val
#
# def fit_peak_cluster_numba_cuda_from_lmfit(zdata, X, Y, lmfit_params, max_nfev=500):
#     param_list = []
#     bounds_min = []
#     bounds_max = []
#
#     n_peaks = max([int(name.split('_')[0][1:]) for name in lmfit_params.keys() if name.startswith('g')]) + 1
#
#     for i in range(n_peaks):
#         for key in ['amplitude', 'radius', 'angle', 'radius_width', 'angle_width', 'theta']:
#             pname = f'g{i}_{key}'
#             p = lmfit_params[pname]
#             param_list.append(p.value)
#             bounds_min.append(-np.inf if p.min is None else p.min)
#             bounds_max.append(np.inf if p.max is None else p.max)
#
#     for key in ['A', 'B', 'C']:
#         p = lmfit_params[key]
#         param_list.append(p.value)
#         bounds_min.append(-np.inf if p.min is None else p.min)
#         bounds_max.append(np.inf if p.max is None else p.max)
#
#     p0 = np.array(param_list)
#     bounds = (np.array(bounds_min), np.array(bounds_max))
#     X_gpu = cuda.to_device(X)
#     Y_gpu = cuda.to_device(Y)
#     zdata_gpu = cuda.to_device(zdata)
#     residuals_gpu = cuda.device_array_like(zdata_gpu)
#
#     def compute_residuals(params):
#         params_gpu = cuda.to_device(params)
#         threadsperblock = 128
#         blockspergrid = (zdata_gpu.size + threadsperblock - 1) // threadsperblock
#         compute_residuals_kernel[blockspergrid, threadsperblock](
#             X_gpu, Y_gpu, zdata_gpu, params_gpu, residuals_gpu, n_peaks
#         )
#         return residuals_gpu.copy_to_host()
#
#     numba_time0 = time.time()
#     result = least_squares(compute_residuals, p0, bounds=bounds, max_nfev=max_nfev)
#     # try:
#     #     result = least_squares(compute_residuals, p0, bounds=bounds, max_nfev=max_nfev)
#     # except:
#     #     result = None
#     numba_time1 = time.time()
#     print("!!!!!!!!!!!! numba_time", (numba_time1 - numba_time0) * 1000)
#     # print(f"numba result{result}")
#
#     return result

############## numba end

############## jaxfit start
# import jax
# import jax.numpy as jnp
# from jaxfit import CurveFit
#
# def two_d_rotated_gaussian_jax(x, y, amp, xo, yo, sigx, sigy, theta):
#     cos_t = jnp.cos(theta)
#     sin_t = jnp.sin(theta)
#     a = (cos_t**2)/(2*sigx**2) + (sin_t**2)/(2*sigy**2)
#     b = -jnp.sin(2*theta)/(4*sigx**2) + jnp.sin(2*theta)/(4*sigy**2)
#     c = (sin_t**2)/(2*sigx**2) + (cos_t**2)/(2*sigy**2)
#     return amp * jnp.exp(-(a*(x-xo)**2 + 2*b*(x-xo)*(y-yo) + c*(y-yo)**2))
#
# def sum_of_gaussians_and_plane_jax(x, y, param_array, n):
#     def add_gaussian(i, z):
#         base = i*6
#         amp = param_array[base]
#         xo = param_array[base+1]
#         yo = param_array[base+2]
#         sigx = param_array[base+3]
#         sigy = param_array[base+4]
#         theta = param_array[base+5]
#
#         z += jax.lax.cond(
#             jnp.any(jnp.isnan(jnp.array([amp, xo, yo, sigx, sigy, theta]))),
#             lambda _: jnp.zeros_like(x),
#             lambda _: two_d_rotated_gaussian_jax(x, y, amp, xo, yo, sigx, sigy, theta),
#             operand=None
#         )
#         return z
#
#     z_init = jnp.zeros_like(x)
#     z = jax.lax.fori_loop(0, n, add_gaussian, z_init)
#
#     a_plane = param_array[n*6] if n*6 < len(param_array) else 0
#     b_plane = param_array[n*6+1] if n*6+1 < len(param_array) else 0
#     c_plane = param_array[n*6+2] if n*6+2 < len(param_array) else 0
#     z += a_plane * x + b_plane * y + c_plane
#     return z
#
# def fit_peak_cluster_jaxfit_from_lmfit(data, X_flat, Y_flat, lm_params):
#     param_names = list(lm_params.keys())
#     n_peaks = sum(1 for name in param_names if name.startswith('g') and '_amplitude' in name)
#
#     print("n_peaks", n_peaks)
#
#     param_list = []
#     bounds_min = []
#     bounds_max = []
#
#     for i in range(n_peaks):
#         for key in ['amplitude', 'radius', 'angle', 'radius_width', 'angle_width', 'theta']:
#             pname = f'g{i}_{key}'
#             p = lm_params[pname]
#             param_list.append(p.value)
#             bounds_min.append(-jnp.inf if p.min is None else p.min)
#             bounds_max.append(jnp.inf if p.max is None else p.max)
#
#     for pname in ['A', 'B', 'C']:
#         p = lm_params[pname]
#         param_list.append(p.value)
#         bounds_min.append(-jnp.inf if p.min is None else p.min)
#         bounds_max.append(jnp.inf if p.max is None else p.max)
#
#     params_jax = jnp.array(param_list, dtype=jnp.float32)
#     print("params_jax", params_jax)
#     bounds = (jnp.array(bounds_min, dtype=jnp.float32), jnp.array(bounds_max, dtype=jnp.float32))
#
#     coords_tuple = [jnp.array(X_flat), jnp.array(Y_flat)]
#     jcf = CurveFit()
#
#     def model_for_fit(coords, *param_array):
#         param_array = jnp.array(param_array)
#         return sum_of_gaussians_and_plane_jax(coords[0], coords[1], param_array, n_peaks).ravel()
#
#     jaxfit_time0 = time.time()
#     try:
#         popt, pcov = jcf.curve_fit(model_for_fit, coords_tuple, data, p0=params_jax, bounds=bounds)
#     except:
#         popt, pcov = None, None
#     jaxfit_time1 = time.time()
#     print("jaxfit_time", (jaxfit_time1 - jaxfit_time0) * 1000)
#     print("popt", popt)
#     return popt, pcov

############## jaxfit end




import pytest
import numpy as np
from pygidfit import ProcessDataFromFile
from pygid import NexusFile


@pytest.mark.parametrize(
    "filename, entry, frame_num, theta_fixed",
    [
        ('./example/251210_HATCN_labeled.h5', None, None, True),  # Rings
        ('./example/251210_DIP_labeled.h5', None, None, True),    # Peaks
        ('./example/BA2PbI4.h5', None, None, True),               # Both
        ('./example/BA2PbI4.h5', 'entry_0001', None, True),       # Single entry
        ('./example/BA2PbI4.h5', 'entry_0000', 1, True),          # Single frame
        ('./example/251210_DIP_labeled.h5', None, None, False),   # Unfixed theta
    ]
)
def test_process_data(filename, entry, frame_num, theta_fixed):
    """
    Integration test for ProcessDataFromFile on different datasets.
    Checks that the class runs without error and outputs expected datasets.
    """
    analysis = ProcessDataFromFile(
        filename,
        entry=entry,
        frame_num=frame_num,
        crit_angle=2,
        clustering_distance_rings=10,
        clustering_distance_peaks=10,
        clustering_extend=2,
        use_pool=False,
        debug=False,
        theta_fixed=theta_fixed,
    )

    # Verify entry_dict exists
    assert hasattr(analysis, 'entry_dict'), "ProcessDataFromFile missing entry_dict attribute"

    nexus = NexusFile(filename)

    entries_to_check = [entry] if entry else list(analysis.entry_dict.keys())

    for e in entries_to_check:
        frames_to_check = [frame_num] if frame_num is not None else range(analysis.entry_dict[e]['shape'][0])
        for f in frames_to_check:
            _check_single_image(nexus, e, f, theta_fixed)


def _check_single_image(nexus, entry, frame, theta_fixed):
    """
    Helper function to check a single frame's fitted peaks.
    """
    # Check required datasets exist
    fitted_peaks = nexus.get_dataset(f'{entry}/data/analysis/frame{frame:05d}/fitted_peaks')
    fitted_peaks_errors = nexus.get_dataset(f'{entry}/data/analysis/frame{frame:05d}/fitted_peaks_errors')
    assert fitted_peaks is not None, f"Fitted peaks missing for {entry} frame {frame}"
    assert fitted_peaks_errors is not None, f"Fitted peaks errors missing for {entry} frame {frame}"

    # Validate lengths match detected peaks
    detected_peaks = nexus.get_dataset(f'{entry}/data/analysis/frame{frame:05d}/detected_peaks')
    assert len(detected_peaks['amplitude']) == len(fitted_peaks['amplitude']), \
        f"Number of fitted peaks does not match detected peaks for {entry} frame {frame}"
    assert len(detected_peaks['amplitude']) == len(fitted_peaks_errors['amplitude']), \
        f"Number of fitted peak errors does not match detected peaks for {entry} frame {frame}"

    # Validate theta values
    if np.isfinite(fitted_peaks['theta']).any():
        if theta_fixed:
            assert np.nanmax(fitted_peaks['theta']) == 0, f"Theta is not fixed for {entry} frame {frame}"
        else:
            assert np.nanmax(fitted_peaks['theta']) != 0, f"Theta should not be fixed for {entry} frame {frame}"

    # Cleanup datasets to avoid side effects
    nexus.delete_dataset(f'{entry}/data/analysis/frame{frame:05d}/fitted_peaks')
    nexus.delete_dataset(f'{entry}/data/analysis/frame{frame:05d}/fitted_peaks_errors')


import numpy as np
from dataclasses import dataclass
from scipy.spatial import cKDTree
from scipy.sparse import coo_matrix
from scipy.sparse.csgraph import connected_components
from shapely.geometry import box as shapely_box
from shapely.strtree import STRtree
import networkx as nx
from typing import List

@dataclass
class Cluster:
    bbox: np.ndarray      # [xmin, ymin, xmax, ymax] - bounding box of the cluster
    bbox_length: int     #  (xmax - xmin)*(ymax-xmin) for peaks and (xmax - xmin) for rings - size of the cluster
    indices: np.ndarray   # (M,) - indices of the original boxes that belong to this cluster
    type: str = None      # 'rings' / 'peaks' / 'both'
    mask_boxes: List = None     # indeces of the pixels to mask

def cluster_peaks(peak_indices, ring_indices, centers, boxes, is_ring_list, r, extend):
    final_clusters = []

    if len(peak_indices) == 0:
        return []

    peak_boxes = boxes[peak_indices]
    peak_geoms = [
        shapely_box(xmin, ymin, xmax, ymax)
        for xmin, ymin, xmax, ymax in peak_boxes
    ]
    tree = STRtree(peak_geoms)
    G = nx.Graph()
    for idx, geom in enumerate(peak_geoms):
        G.add_node(idx)
        hits = tree.query(geom.buffer(r))
        for j in hits:
            if idx != j:
                G.add_edge(idx, j)

    components = list(nx.connected_components(G))

    for component in components:
        local_peak_idx = np.array(list(component))
        cluster_peak_indices = peak_indices[local_peak_idx]
        cluster_boxes = boxes[cluster_peak_indices]
        cluster_centers = centers[cluster_peak_indices]

        xmin = np.min(cluster_boxes[:, 0]) - extend
        ymin = np.min(cluster_boxes[:, 1]) - extend
        xmax = np.max(cluster_boxes[:, 2]) + extend
        ymax = np.max(cluster_boxes[:, 3]) + extend

        if xmax < xmin or ymax < ymin:
            print("xmin, ymin, xmax, ymax", xmin, ymin, xmax, ymax)

        # h_box = min((ymax - ymin) * 0.5, r * 0.3)
        # w_box = min((xmax - xmin), r * 0.3)
        # xmin -= w_box
        # ymin -= h_box
        # xmax += w_box
        # ymax += h_box

        bbox = np.array([xmin, ymin, xmax, ymax])

        cluster_ring_indices = []
        if len(ring_indices) > 0:
            cluster_x = centers[cluster_peak_indices][:, 0]
            for i in ring_indices:
                ring_x = centers[i, 0]
                if np.any(np.abs(cluster_x - ring_x) <= r):
                    cluster_ring_indices.append(i)

        cluster_ring_indices = np.array(cluster_ring_indices, dtype=int)
        all_indices = np.concatenate([cluster_peak_indices, cluster_ring_indices])
        all_is_ring = is_ring_list[all_indices]

        mask_boxes = []

        for box, ind  in zip(peak_boxes, peak_indices):
            if ind in cluster_peak_indices:
                continue

            xmin_current, ymin_current, xmax_current, ymax_current = box

            intersects = not (xmax_current < xmin or xmin_current > xmax or
                              ymax_current < ymin or ymin_current > ymax)
            if intersects:
                xmin_mask = max(xmin_current, xmin)
                ymin_mask = max(ymin_current, ymin)
                xmax_mask = min(xmax_current, xmax)
                ymax_mask = min(ymax_current, ymax)
                mask_boxes.append((xmin_mask, ymin_mask, xmax_mask, ymax_mask))

        # if len(mask_boxes) != 0:
        #     print("FOUND TO MASK")
        #
        #     import matplotlib.pyplot as plt
        #     import matplotlib.patches as patches
        #     fig, ax = plt.subplots()
        #     cluster_rect = patches.Rectangle(
        #         (xmin, ymin),
        #         xmax - xmin,
        #         ymax - ymin,
        #         linewidth=2,
        #         edgecolor='green',
        #         facecolor='none',
        #         label='Cluster'
        #     )
        #     ax.add_patch(cluster_rect)
        #
        #     for box in peak_boxes:
        #         rect = patches.Rectangle(
        #             (box[0], box[1]),
        #             box[2] - box[0],
        #             box[3] - box[1],
        #             linewidth=1,
        #             edgecolor='blue',
        #             facecolor='none',
        #             linestyle='--'
        #         )
        #         ax.add_patch(rect)
        #
        #     for box in mask_boxes:
        #         rect = patches.Rectangle(
        #             (box[0], box[1]),
        #             box[2] - box[0],
        #             box[3] - box[1],
        #             linewidth=2,
        #             edgecolor='red',
        #             facecolor='none',
        #             label='Mask'
        #         )
        #         ax.add_patch(rect)
        #     ax.set_xlim(xmin-10, xmax+10)
        #     ax.set_ylim(ymin-10, ymax+10)
        #     ax.set_aspect('equal')
        #     ax.set_xlabel(str(len(mask_boxes)) + str(cluster_peak_indices))
        #     plt.legend()
        #     plt.show()


        final_clusters.append(Cluster(
            bbox=bbox,
            bbox_length = int((xmax-xmin)*(ymax-ymin)),
            indices=all_indices,
            type='both' if True in all_is_ring else 'peaks',
            mask_boxes = mask_boxes
        ))

    return final_clusters


def cluster_peaks_centers(peak_indices, ring_indices, centers, boxes, is_ring_list, r, extend):
    final_clusters = []

    if len(peak_indices) == 0:
        return []

    peak_centers = centers[peak_indices]
    tree_peaks = cKDTree(peak_centers)

    pairs = tree_peaks.query_pairs(r=r, p=np.inf)
    if pairs:
        i, j = zip(*pairs)
        row = np.concatenate([i, j])
        col = np.concatenate([j, i])
        data = np.ones(len(row), dtype=bool)
        adj = coo_matrix((data, (row, col)), shape=(len(peak_indices), len(peak_indices)))
        _, peak_labels = connected_components(adj, directed=False)
    else:
        peak_labels = np.arange(len(peak_indices))

    for cluster_id in np.unique(peak_labels):
        local_peak_idx = np.where(peak_labels == cluster_id)[0]
        cluster_peak_indices = peak_indices[local_peak_idx]
        cluster_boxes = boxes[cluster_peak_indices]
        cluster_centers = centers[cluster_peak_indices]

        xmin = np.min(cluster_boxes[:, 0]) - extend
        ymin = np.min(cluster_boxes[:, 1]) - extend
        xmax = np.max(cluster_boxes[:, 2]) + extend
        ymax = np.max(cluster_boxes[:, 3]) + extend

        h_box = min((ymax - ymin) * 0.5, r * 0.3)
        w_box = min((xmax - xmin), r * 0.3)
        xmin -= w_box
        ymin -= h_box
        xmax += w_box
        ymax += h_box

        bbox = np.array([xmin, ymin, xmax, ymax])

        cluster_ring_indices = []
        if len(ring_indices) > 0:
            cluster_x = centers[cluster_peak_indices][:, 0]
            for i in ring_indices:
                ring_x = centers[i, 0]
                if np.any(np.abs(cluster_x - ring_x) <= r):
                    cluster_ring_indices.append(i)

        cluster_ring_indices = np.array(cluster_ring_indices, dtype=int)
        all_indices = np.concatenate([cluster_peak_indices, cluster_ring_indices])
        all_is_ring = is_ring_list[all_indices]


        final_clusters.append(Cluster(
            bbox=bbox,
            bbox_length=(xmax - xmin) * (ymax - ymin),
            indices=all_indices,
            type='both' if True in all_is_ring else 'peaks'
        ))

    return final_clusters


def cluster_rings(ring_indices, centers, boxes, r, extend):
    final_clusters = []

    if len(ring_indices) == 0:
        return []

    ring_centers = centers[ring_indices]
    x_coords = ring_centers[:, 0].reshape(-1, 1)

    tree = cKDTree(x_coords)
    pairs = tree.query_pairs(r=r, p=np.inf)

    if pairs:
        i, j = zip(*pairs)
        row = np.concatenate([i, j])
        col = np.concatenate([j, i])
        data = np.ones(len(row), dtype=bool)
        adj = coo_matrix((data, (row, col)), shape=(len(ring_indices), len(ring_indices)))
        _, labels = connected_components(adj, directed=False)
    else:
        labels = np.arange(len(ring_indices))

    for cluster_id in np.unique(labels):
        local_idx = np.where(labels == cluster_id)[0]
        cluster_ring_indices = np.array(ring_indices)[local_idx]
        cluster_boxes = boxes[cluster_ring_indices]

        xmin = np.min(cluster_boxes[:, 0]) - extend
        ymin = np.min(cluster_boxes[:, 1]) - extend
        xmax = np.max(cluster_boxes[:, 2]) + extend
        ymax = np.max(cluster_boxes[:, 3]) + extend

        w_box = min((xmax - xmin), r * 0.2)
        xmin -= w_box
        xmax += w_box

        bbox = np.array([xmin, ymin, xmax, ymax])
        final_clusters.append(Cluster(
            bbox=bbox,
            bbox_length=int((xmax - xmin)),
            indices=cluster_ring_indices,
            type='rings',
            mask_boxes = [],
        ))

    return final_clusters

def cluster_boxes_by_centers(boxes_list, r_peaks, r_rings, extend):
    boxes_id = np.array([box.index for box in boxes_list])
    is_ring_list = np.array([box.is_ring for box in boxes_list], dtype=bool)

    boxes = np.stack([box.limits for box in boxes_list])
    centers = 0.5 * (boxes[:, :2] + boxes[:, 2:])

    peak_indices = np.where(~is_ring_list)[0]
    ring_indices = np.where(is_ring_list)[0]

    clusters = []
    clusters += cluster_peaks(peak_indices, ring_indices, centers, boxes, is_ring_list, r_peaks, extend)
    clusters += cluster_rings(ring_indices, centers, boxes, r_rings, extend)
    return clusters



import time
from pathlib import Path
import numpy as np

from pygidfit.process_scans import process_data_from_file, process_data_img_container



def run_scans(
        filename=None,
        entry = None,
        frame_num = None,
        batch_size=10,
        crit_angle=0,
        polar_shape=np.array([512, 1024]),
        clustering_distance_peaks=7,
        clustering_distance_rings=40,
        clustering_extend=2,
        use_pool=False,
        debug=False,
        multiprocessing=True,
):
    """
    Main function to process a range of scans.

    Args:
        filename: Name of the file
        scan_range: Range of scan numbers to process (start, end)
        box_shape: Shape of the box (width, height)
        n_jobs: Number of parallel jobs to use (-1 for all cores)
    """

    process_data_from_file(
        filename,
        entry,
        frame_num,
        batch_size,
        crit_angle,
        polar_shape,
        clustering_distance_peaks,
        clustering_distance_rings,
        clustering_extend,
        use_pool,
        debug,
        multiprocessing
    )


def run_scans_img_container(
        img_container=None,
        batch_size=10,
        crit_angle=0,
        polar_shape=np.array([512, 1024]),
        clustering_distance_peaks=7,
        clustering_distance_rings=40,
        clustering_extend=2,
        use_pool=False,
        debug=False,
        multiprocessing=False,
):
    """
    Main function to process a range of scans.

    Args:
        filename: Name of the file
        scan_range: Range of scan numbers to process (start, end)
        box_shape: Shape of the box (width, height)
        n_jobs: Number of parallel jobs to use (-1 for all cores)
    """

    return process_data_img_container(
        img_container,
        crit_angle,
        polar_shape,
        clustering_distance_peaks,
        clustering_distance_rings,
        clustering_extend,
        use_pool,
        debug,
        multiprocessing
    )


# filename = 'S124_FAI_A1_after_gui_copy_with_duplicate.h5'
if __name__ == "__main__":
    filename = r'/tests/240305_PEN_DIP_0001.h5'
    run_scans(filename,
              polar_shape=[512 * 2, 1024 * 2],
              batch_size=1,
              crit_angle=0.1,
              debug=False,
              multiprocessing=False)



import numpy as np
from dataclasses import dataclass

@dataclass
class Boxes:
    limits: np.ndarray
    is_ring: bool
    index: int
    # cluster_num: List[int] = field(default_factory=list)
    is_cut_qz: bool = False
    is_cut_qxy: bool = False
    fitting_result = None


def find_box_type(is_cut_qxy, is_cut_qz, box, q, q_xy_max, q_z_max, polar_shape_0) -> bool:
    """
    Classifies a bounding box as either a 'line' or a 'peak' based on its aspect ratio.

    Args:
        box (array-like): [xmin, ymin, xmax, ymax]

    Returns:
        str: 'line' or 'peak'
    """
    result1 = True if (is_cut_qz and is_cut_qxy) else False

    xmin, ymin, xmax, ymax = box
    width = xmax - xmin
    height = ymax - ymin

    ang_hor = 0 if q < q_xy_max else np.arccos(q_xy_max/q)
    ang_vert = 0 if q < q_z_max else np.arccos(q_z_max/q)
    ratio = (np.pi/2 - ang_hor - ang_vert) / (np.pi/2)
    result3 = height > 0.6 * polar_shape_0 * ratio
    # if result3:
    #     print("found ring: ", np.rad2deg(ang_hor), np.rad2deg(ang_vert), ratio, xmin, ymin, xmax, ymax)
    return result1 or result3

    #
    # return True if shape <= height*1.5 else False




def boxes_preprocessing(detected_peaks, polar_shape, wavelength, q_abs_max,
                        q_xy_max, q_z_max):

    radius1_q = detected_peaks.radius - (detected_peaks.radius_width / 1) #2
    radius2_q = detected_peaks.radius + (detected_peaks.radius_width / 1) #2


    radius1 = np.round(radius1_q / q_abs_max * polar_shape[1]) - 1
    radius2 = np.round(radius2_q / q_abs_max * polar_shape[1]) + 1

    theta1_deg = detected_peaks.angle - (np.abs(detected_peaks.angle_width) / 2)
    theta2_deg = detected_peaks.angle + (np.abs(detected_peaks.angle_width) / 2)
    theta1 = np.round(theta1_deg / 90 * polar_shape[0]).astype(int) - 1
    theta2 = np.round(theta2_deg / 90 * polar_shape[0]).astype(int) + 1

    # to make boxes inside the img
    radius1 = np.clip(radius1, 0, polar_shape[1])
    radius2 = np.clip(radius2, 0, polar_shape[1])
    theta1 = np.clip(theta1, 0, polar_shape[0])
    theta2 = np.clip(theta2, 0, polar_shape[0])

    for i in range(len(theta1)):
        if theta1[i] == theta2[i]:
            print(theta1[i], theta2[i], detected_peaks.angle[i], detected_peaks.angle_width[i])
            raise


    boxes = np.stack([radius1,
                           theta1,
                           radius2,
                           theta2], axis=1)
    x0c = (radius1+radius2)/2
    y0c = (theta1+theta2)/2

    boxes_q_deg = np.stack([radius1_q,
                           theta1_deg,
                           radius2_q,
                           theta2_deg], axis=1)
    boxes_list = []
    for i in range(len(radius1)):
        is_cut_qxy, is_cut_qz = _get_cut_flags(radius1_q[i],theta1_deg[i],radius2_q[i],theta2_deg[i],wavelength)
        boxes_list.append(Boxes(boxes[i],
                                find_box_type(is_cut_qxy, is_cut_qz, boxes[i],
                                              detected_peaks.radius[i], q_xy_max, q_z_max, polar_shape[0]),
                                i,is_cut_qxy, is_cut_qz ))
        boxes_list[i].boxes_q_deg = boxes_q_deg[i]
        boxes_list[i].x0c = x0c[i]
        boxes_list[i].y0c = y0c[i]




    return boxes_list

def _get_cut_flags(radius1, theta1_deg, radius2, theta2_deg, wavelength):
    if theta1_deg < 2:
        is_cut_qxy = True
    else:
        is_cut_qxy = False
    is_cut_qz = _get_missing_wedge_pol(wavelength, radius2, theta2_deg)
    return is_cut_qz, is_cut_qxy

def _get_missing_wedge_pol(wavelength, q_abs, phi):
    k = 2 * np.pi / float(wavelength)
    return bool(np.abs(q_abs) > np.abs(2*k*np.cos(np.deg2rad(phi))))


def make_box_attributes(indices, boxes, fitting_result, type = None, debug = False):

    params = fitting_result['params']
    errors = fitting_result['errors']

    i = 0
    for ind in indices:
        box = boxes[ind]
        if box.is_ring and box.fitting_result is not None:
            continue
        box.fitting_result = {}
        for key in params.keys():
            if key.startswith(f'g{i}_'):
                key_changed = key.replace(f'g{i}_', '')
                box.fitting_result[key_changed] = params[key]
            elif key in ['A', 'B', 'C']:
                box.fitting_result[key] = params[key]
            elif key == 'lin_slope':
                box.fitting_result['A'] = params[key]
            elif key == 'lin_intercept':
                box.fitting_result['C'] = params[key]

        if not 'B' in box.fitting_result:
            box.fitting_result['B'] = 0
        if not 'angle' in box.fitting_result:
            box.fitting_result['angle'] = np.nan
        if not 'angle_width' in box.fitting_result:
            box.fitting_result['angle_width'] = np.inf
        if not 'theta' in box.fitting_result:
            box.fitting_result['theta'] = np.nan

        box.fitting_error = {}
        for key in errors.keys():
            if key.startswith(f'g{i}_'):
                key_changed = key.replace(f'g{i}_', '')
                box.fitting_error[key_changed] = errors[key]
            elif key in ['A', 'B', 'C']:
                box.fitting_error[key] = errors[key]
            elif key == 'lin_slope':
                box.fitting_error['A'] = errors[key]
            elif key == 'lin_intercept':
                box.fitting_error['C'] = errors[key]

        if not 'B' in box.fitting_error:
            box.fitting_error['B'] = 0
        if not 'angle' in box.fitting_error:
            box.fitting_error['angle'] = np.nan
        if not 'angle_width' in box.fitting_error:
            box.fitting_error['angle_width'] = np.inf
        if not 'theta' in box.fitting_error:
            box.fitting_error['theta'] = np.nan
        i+=1
        if debug:
            print("fitting_result",box.fitting_result)
            print("fitting_error", box.fitting_error)
